from __future__ import print_function, division, absolute_import
from fontTools.misc.py23 import *
from fontTools import ttLib

superclass = ttLib.getTableClass("fpgm")

class table__p_r_e_p(superclass):
	pass
